@@ -4,11 +4,13 @@ use serenity::async_trait;
 use serenity::model::prelude::*;
 use serenity::prelude::*;
 use std::fmt::Debug;
-use tokio::sync::Mutex;
+use std::ops::DerefMut;
 use std::str::FromStr;
+use tokio::sync::Mutex;
 
 struct Handler {
     prefix: String,
+    connector: Box<dyn Connector + Send + Sync>,
     command: Option<String>,
     connection: Mutex<Connection>,
     role: Option<RoleId>,
@@ -19,45 +21,63 @@ struct Handler {
 impl EventHandler for Handler {
     async fn message(&self, ctx: Context, message: Message) {
         if message.author.bot {
-            return
+            return;
         }
         let permission = true;
-        let permission = permission && if let Some(channel) = self.channel {
-            message.channel_id == channel
-        } else {
-            true
-        };
-        let permission = permission && if let Some(guild) = message.guild_id {
-            if let Some(expected_role) = self.role {
-                message.author.has_role(&ctx.http, guild, expected_role).await.unwrap_or(false)
+        let permission = permission
+            && if let Some(channel) = self.channel {
+                message.channel_id == channel
             } else {
                 true
-            }
-        } else {
-            false
+            };
+        let permission = permission
+            && if let Some(guild) = message.guild_id {
+                if let Some(expected_role) = self.role {
+                    message
+                        .author
+                        .has_role(&ctx.http, guild, expected_role)
+                        .await
+                        .unwrap_or(false)
+                } else {
+                    true
+                }
+            } else {
+                false
+            };
+        if !permission {
+            return;
         };
-        if !permission { 
-            return
-        };
-        if let Some(command) = self.command.as_ref().and_then(|pat| message.content.strip_prefix(pat)) {
-            match command.split_once(char::is_whitespace).map(|x| x.0) {
-                Some("stop") => {
-                    print_err(
-                        message
-                            .channel_id
-                            .say(&ctx.http, "stopping ...")
-                            .await,
-                    );
+        if let Some(command) = self
+            .command
+            .as_ref()
+            .and_then(|pat| message.content.strip_prefix(pat))
+        {
+            let command = command.trim();
+            let mut iterator = command.split_whitespace();
+            match iterator.next() {
+                Some("reconnect") => {
+                    print_err(message.channel_id.say(&ctx.http, "reconnecting ...").await);
+                    match self.connector.connect().await {
+                        Ok(connection) => {
+                            *self.connection.lock().await.deref_mut() = connection;
+                            print_err(message.channel_id.say(&ctx.http, "reconnected").await);
+                        }
+                        Err(err) => {
+                            print_err(
+                                message
+                                    .channel_id
+                                    .say(&ctx.http, format!("error connecting to server: {}\nold connection still used.", err))
+                                    .await,
+                            );
+                        }
+                    }
                     panic!("manually stop")
                 }
                 _ => {
                     print_err(
                         message
                             .channel_id
-                            .say(
-                                &ctx.http,
-                                format!("unknown command: {}", command),
-                            )
+                            .say(&ctx.http, format!("unknown command: {}", command))
                             .await,
                     );
                 }
@@ -66,9 +86,7 @@ impl EventHandler for Handler {
             let command = command.trim();
             println!("run command: {}", command);
 
-            match {
-                (*self.connection.lock().await).cmd(command).await 
-            } {
+            match { (*self.connection.lock().await).cmd(command).await } {
                 Ok(res) => {
                     println!("response: {}", res);
                     print_err(
@@ -110,23 +128,19 @@ async fn main() {
     // Discord Bot Token を設定
     let options = read_options().await;
 
-    let mut builder = Connection::builder();
-    match options.server_kind {
-        ServerKind::Normal => {}
-        ServerKind::Minecraft => builder = builder.enable_minecraft_quirks(true),
-        ServerKind::Factorio => builder = builder.enable_factorio_quirks(true),
-    }
-    let connection = builder
-        .connect(options.rcon.address, &options.rcon.pass)
-        .await
-        .expect("failed to connect");
+    let connector = SimpleConnector {
+        server_kind: options.server_kind,
+        address: options.rcon.address.clone(),
+        pass: options.rcon.pass.clone(),
+    };
 
     let handler = Handler {
         prefix: options.prefix,
         command: options.command,
         role: options.role.map(RoleId),
         channel: options.channel.map(ChannelId),
-        connection: Mutex::new(connection),
+        connection: Mutex::new(connector.connect().await.expect("failed to connect")),
+        connector: Box::new(connector),
     };
 
     let mut client = Client::builder(&options.token)
@@ -142,36 +156,57 @@ async fn main() {
 }
 
 async fn read_options() -> Options {
-    if let Some(options) = read_options_from_env() {
-        return options
+    let expect_env = match read_options_from_env() {
+        Ok(options) => return options,
+        Err(expect_env) => expect_env,
+    };
+    let err_msg = match read_options_from_file().await {
+        Ok(options) => return options,
+        Err(err_msg) => err_msg,
+    };
+    let mut args = std::env::args();
+    args.next(); // skip exec name
+    let prefer_env = args.next().map(|x| &x == "--env").unwrap_or(false);
+    if prefer_env {
+        panic!("required envionemnt variable {} not found", expect_env);
+    } else {
+        panic!("{}", err_msg);
     }
-    let config = tokio::fs::read_to_string("config.toml")
-        .await
-        .expect("failed to read config.toml");
-    toml::from_str(&config).expect("failed to read config.toml")
 }
 
-fn read_options_from_env() -> Option<Options> {
+fn read_options_from_env() -> Result<Options, &'static str> {
     use std::env;
 
-    Some(Options {
-        token: env::var("DISCORD_TOKEN").ok()?,
-        prefix: env::var("DISCORD_PREFIX").ok()?,
+    Ok(Options {
+        token: env::var("DISCORD_TOKEN").map_err(|_| "DISCORD_TOKEN")?,
+        prefix: env::var("DISCORD_PREFIX").map_err(|_| "DISCORD_PREFIX")?,
         command: env::var("DISCORD_COMMAND").ok(),
-        role: env::var("DISCORD_ROLE").ok()
+        role: env::var("DISCORD_ROLE")
+            .ok()
             .map(|x| x.parse::<u64>().expect("parsing DISCORD_ROLE")),
-        channel: env::var("DISCORD_CHANNEL").ok()
+        channel: env::var("DISCORD_CHANNEL")
+            .ok()
             .map(|x| x.parse::<u64>().expect("parsing DISCORD_CHANNEL")),
-        server_kind: env::var("SERVER_KIND").ok()
+        server_kind: env::var("SERVER_KIND")
+            .ok()
             .map(|x| x.parse::<ServerKind>().expect("parsing SERVER_KIND"))
             .unwrap_or_else(Default::default),
         rcon: RconOptions {
-            address: env::var("RCON_ADDRESS").ok()?,
-            pass: env::var("RCON_PASSWORD").ok().unwrap_or_else(Default::default),
-        }
+            address: env::var("RCON_ADDRESS").map_err(|_| "RCON_ADDRESS")?,
+            pass: env::var("RCON_PASSWORD")
+                .ok()
+                .unwrap_or_else(Default::default),
+        },
     })
 }
 
+async fn read_options_from_file() -> Result<Options, &'static str> {
+    let config = tokio::fs::read_to_string("config.toml")
+        .await
+        .map_err(|_| "failed to read config.toml")?;
+    Ok(toml::from_str(&config).expect("failed to read config.toml"))
+}
+
 #[derive(Deserialize)]
 struct Options {
     token: String,
@@ -191,7 +226,7 @@ struct RconOptions {
     pass: String,
 }
 
-#[derive(Deserialize)]
+#[derive(Deserialize, Clone, Copy)]
 enum ServerKind {
     #[serde(rename = "normal")]
     Normal,
@@ -212,7 +247,10 @@ struct ParseServerKindErr(());
 
 impl std::fmt::Display for ParseServerKindErr {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "value is invalid. oneof `normal`, `minecraft`, or `factorio` is required.")
+        write!(
+            f,
+            "value is invalid. oneof `normal`, `minecraft`, or `factorio` is required."
+        )
     }
 }
 
@@ -230,3 +268,27 @@ impl FromStr for ServerKind {
         }
     }
 }
+
+struct SimpleConnector {
+    server_kind: ServerKind,
+    address: String,
+    pass: String,
+}
+
+#[async_trait]
+impl Connector for SimpleConnector {
+    async fn connect(&self) -> rcon::Result<Connection> {
+        let mut builder = Connection::builder();
+        match self.server_kind {
+            ServerKind::Normal => {}
+            ServerKind::Minecraft => builder = builder.enable_minecraft_quirks(true),
+            ServerKind::Factorio => builder = builder.enable_factorio_quirks(true),
+        }
+        builder.connect(&self.address, &self.pass).await
+    }
+}
+
+#[async_trait]
+trait Connector {
+    async fn connect(&self) -> rcon::Result<Connection>;
+}
